import inspect
from base64 import b32encode
from collections.abc import MutableMapping
from copy import copy
from enum import Enum, IntEnum
from hashlib import sha3_224
from typing import TYPE_CHECKING, Any, Optional, Type, Union

from pydantic import ConfigDict
from pydantic import Field as PydanticField
from pydantic import computed_field, create_model

from tortoise import (
    BackwardFKRelation,
    BackwardOneToOneRelation,
    ForeignKeyFieldInstance,
    ManyToManyFieldInstance,
    OneToOneFieldInstance,
)
from tortoise.contrib.pydantic.base import PydanticModel
from tortoise.contrib.pydantic.descriptions import (
    ComputedFieldDescription,
    ModelDescription,
    PydanticMetaData,
)
from tortoise.contrib.pydantic.utils import get_annotations
from tortoise.fields import Field, JSONField
from tortoise.fields.data import CharEnumFieldInstance, IntEnumFieldInstance

if TYPE_CHECKING:  # pragma: nocoverage
    from tortoise.models import Model

_MODEL_INDEX: dict[str, Type[PydanticModel]] = {}


def _br_it(val: str) -> str:
    return val.replace("\n", "<br/>").strip()


def _cleandoc(obj: Any) -> str:
    return _br_it(inspect.cleandoc(obj.__doc__ or ""))


def _pydantic_recursion_protector(
        cls: "Type[Model]",
        *,
        stack: tuple,
        exclude: tuple[str, ...] = (),
        include: tuple[str, ...] = (),
        computed: tuple[str, ...] = (),
        name=None,
        allow_cycles: bool = False,
        sort_alphabetically: Optional[bool] = None,
) -> Optional[Type[PydanticModel]]:
    """
    It is an inner function to protect pydantic model creator against cyclic recursion
    """
    if not allow_cycles and cls in (c[0] for c in stack[:-1]):
        return None

    caller_fname = stack[0][1]
    prop_path = [caller_fname]  # It stores the fields in the hierarchy
    level = 1
    for _, parent_fname, parent_max_recursion in stack[1:]:
        # Check recursion level
        prop_path.insert(0, parent_fname)
        if level >= parent_max_recursion:
            # This is too verbose, Do we even need a way of reporting truncated models?
            # tortoise.logger.warning(
            #     "Recursion level %i has reached for model %s",
            #     level,
            #     parent_cls.__qualname__ + "." + ".".join(prop_path),
            # )
            return None

        level += 1
    pmc = PydanticModelCreator(
        cls,
        exclude=exclude,
        include=include,
        computed=computed,
        name=name,
        _stack=stack,
        allow_cycles=allow_cycles,
        sort_alphabetically=sort_alphabetically,
        _as_submodel=True,
    )
    return pmc.create_pydantic_model()


class FieldMap(MutableMapping[str, Union[Field, ComputedFieldDescription]]):
    def __init__(self, meta: PydanticMetaData, pk_field: Optional[Field] = None):
        self._field_map: dict[str, Union[Field, ComputedFieldDescription]] = {}
        self.pk_raw_field = pk_field.model_field_name if pk_field is not None else ""
        if pk_field:
            self.pk_raw_field = pk_field.model_field_name
            self.field_map_update([pk_field], meta)
        self.computed_fields: dict[str, ComputedFieldDescription] = {}

    def __delitem__(self, __key):
        self._field_map.__delitem__(__key)

    def __getitem__(self, __key):
        return self._field_map.__getitem__(__key)

    def __len__(self):  # pragma: no-coverage
        return self._field_map.__len__()

    def __iter__(self):
        return self._field_map.__iter__()

    def __setitem__(self, __key, __value):
        self._field_map.__setitem__(__key, __value)

    def sort_alphabetically(self) -> None:
        self._field_map = {k: self._field_map[k] for k in sorted(self._field_map)}

    def sort_definition_order(self, cls: "Type[Model]", computed: tuple[str, ...]) -> None:
        self._field_map = {
            k: self._field_map[k]
            for k in tuple(cls._meta.fields_map.keys()) + computed
            if k in self._field_map
        }

    def field_map_update(self, fields: list[Field], meta: PydanticMetaData) -> None:
        for field in fields:
            name = field.model_field_name
            # Include or exclude field
            if (meta.include and name not in meta.include) or name in meta.exclude:
                continue
            # Remove raw fields
            if isinstance(field, ForeignKeyFieldInstance):
                raw_field = field.source_field
                if (
                        raw_field is not None
                        and meta.exclude_raw_fields
                        and raw_field != self.pk_raw_field
                ):
                    self.pop(raw_field, None)
            self[name] = field

    def computed_field_map_update(self, computed: tuple[str, ...], cls: "Type[Model]"):
        self._field_map.update(
            {
                k: ComputedFieldDescription(
                    field_type=callable,
                    function=getattr(cls, k),
                    description=None,
                )
                for k in computed
            }
        )


class PydanticModelCreator:
    def __init__(
            self,
            cls: "Type[Model]",
            name: Optional[str] = None,
            exclude: Optional[tuple[str, ...]] = None,
            include: Optional[tuple[str, ...]] = None,
            computed: Optional[tuple[str, ...]] = None,
            optional: Optional[tuple[str, ...]] = None,
            allow_cycles: Optional[bool] = None,
            sort_alphabetically: Optional[bool] = None,
            exclude_readonly: bool = False,
            meta_override: Optional[Type] = None,
            model_config: Optional[ConfigDict] = None,
            validators: Optional[dict[str, Any]] = None,
            module: str = __name__,
            _stack: tuple = (),
            _as_submodel: bool = False,
    ) -> None:
        self._cls: "Type[Model]" = cls
        self._stack: tuple[tuple["Type[Model]", str, int], ...] = (
            _stack  # ((Type[Model], field_name, max_recursion),)
        )
        self._is_default: bool = (
                exclude is None
                and include is None
                and computed is None
                and optional is None
                and sort_alphabetically is None
                and allow_cycles is None
                and meta_override is None
                and not exclude_readonly
        )
        if exclude is None:
            exclude = ()
        if include is None:
            include = ()
        if computed is None:
            computed = ()
        if optional is None:
            optional = ()

        if meta := getattr(cls, "PydanticMeta", None):
            meta_from_class = PydanticMetaData.from_pydantic_meta(meta)
        else:  # default
            meta_from_class = PydanticMetaData()
        if meta_override:
            meta_from_class = meta_from_class.construct_pydantic_meta(meta_override)
        self.meta = meta_from_class.finalize_meta(
            exclude=exclude,
            include=include,
            computed=computed,
            allow_cycles=allow_cycles,
            sort_alphabetically=sort_alphabetically,
            model_config=model_config,
        )

        self._exclude_read_only: bool = exclude_readonly

        self._fqname = cls.__module__ + "." + cls.__qualname__
        self._name: str
        self._title: str
        self.given_name = name
        self.__hash: str = ""

        self._as_submodel = _as_submodel

        self._annotations = get_annotations(cls)

        self._pconfig: ConfigDict

        self._properties: dict[str, Any] = dict()
        self._relational_fields_index: list[tuple[str, str]] = list()

        self._model_description: ModelDescription = ModelDescription.from_model(cls)

        self._field_map: FieldMap = self._initialize_field_map()
        self._construct_field_map()

        self._optional = optional

        self._validators = validators
        self._module = module

        self._stack = _stack

    @property
    def _hash(self):
        if self.__hash == "":
            hashval = (
                f"{self._fqname};{self._properties.keys()};{self._relational_fields_index};{self._optional};"
                f"{self.meta.allow_cycles}"
            )
            self.__hash = (
                b32encode(sha3_224(hashval.encode("utf-8")).digest()).decode("utf-8").lower()[:6]
            )
        return self.__hash

    def get_name(self) -> tuple[str, str]:
        # If arguments are specified (different from the defaults), we append a hash to the
        # class name, to make it unique
        # We don't check by stack, as cycles get explicitly renamed.
        # When called later, include is explicitly set, so fence passes.
        if self.given_name is not None:
            return self.given_name, self.given_name
        name = f"{self._fqname}:{self._hash}" if not self._is_default else self._fqname
        name = f"{name}:leaf" if self._as_submodel else name
        return name, self._cls.__name__

    def _initialize_pconfig(self) -> ConfigDict:
        pconfig: ConfigDict = PydanticModel.model_config.copy()
        if self.meta.model_config:
            pconfig.update(self.meta.model_config)
        if "title" not in pconfig:
            pconfig["title"] = self._title
        if "extra" not in pconfig:
            pconfig["extra"] = "forbid"
        return pconfig

    def _initialize_field_map(self) -> FieldMap:
        return (
            FieldMap(self.meta)
            if self._exclude_read_only
            else FieldMap(self.meta, pk_field=self._model_description.pk_field)
        )

    def _construct_field_map(self) -> None:
        self._field_map.field_map_update(fields=self._model_description.data_fields, meta=self.meta)
        if not self._exclude_read_only:
            for fields in (
                    self._model_description.fk_fields,
                    self._model_description.o2o_fields,
                    self._model_description.m2m_fields,
            ):
                self._field_map.field_map_update(fields, self.meta)
            if self.meta.backward_relations:
                for fields in (
                        self._model_description.backward_fk_fields,
                        self._model_description.backward_o2o_fields,
                ):
                    self._field_map.field_map_update(fields, self.meta)
            self._field_map.computed_field_map_update(self.meta.computed, self._cls)
        if self.meta.sort_alphabetically:
            self._field_map.sort_alphabetically()
        else:
            self._field_map.sort_definition_order(self._cls, self.meta.computed)

    def create_pydantic_model(self) -> Type[PydanticModel]:
        for field_name, field in self._field_map.items():
            self._process_field(field_name, field)

        self._name, self._title = self.get_name()

        if self._hash in _MODEL_INDEX:
            # there is a model exactly the same, but the name could be different
            hashed_model = _MODEL_INDEX[self._hash]
            if hashed_model.__name__ == self._name:
                # also the same name
                return _MODEL_INDEX[self._hash]

        self._pconfig = self._initialize_pconfig()
        self._properties["model_config"] = self._pconfig
        model = create_model(
            self._name,
            __base__=PydanticModel,
            __module__=self._module,
            __validators__=self._validators,
            **self._properties,
        )
        # Copy the Model docstring over
        model.__doc__ = _cleandoc(self._cls)
        # Store the base class
        model.model_config["orig_model"] = self._cls  # type: ignore
        # Store model reference so we can de-dup it later on if needed.
        _MODEL_INDEX[self._hash] = model
        return model

    def _process_field(
            self,
            field_name: str,
            field: Union[Field, ComputedFieldDescription],
    ) -> None:
        json_schema_extra: dict[str, Any] = {}
        fconfig: dict[str, Any] = {
            "json_schema_extra": json_schema_extra,
        }
        field_property: Optional[Any] = None
        is_to_one_relation: bool = False
        if isinstance(field, Field):
            field_property, is_to_one_relation = self._process_normal_field(
                field_name, field, json_schema_extra, fconfig
            )
            if field_property:
                fconfig["title"] = field_name.replace("_", " ").title()
                description = _br_it(field.docstring or field.description or "")
                if description:
                    fconfig["description"] = description
                if field_name in self._optional or (
                        field.default is not None and not callable(field.default)
                ):
                    self._properties[field_name] = (
                        field_property,
                        PydanticField(default=field.default, **fconfig),
                    )
                else:
                    # if (json_schema_extra.get("nullable") and not is_to_one_relation) or (
                    #     self._exclude_read_only and json_schema_extra.get("readOnly")
                    # ):
                    #     # see: https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
                    #     fconfig["default"] = None
                    self._properties[field_name] = (field_property, PydanticField(**fconfig, default=None))
        elif isinstance(field, ComputedFieldDescription):
            field_property, is_to_one_relation = self._process_computed_field(field), False
            if field_property:
                comment = _cleandoc(field.function)
                fconfig["title"] = field_name.replace("_", " ").title()
                description = comment or _br_it(field.description or "")
                if description:
                    fconfig["description"] = description
                self._properties[field_name] = field_property

    def _process_normal_field(
            self,
            field_name: str,
            field: Field,
            json_schema_extra: dict[str, Any],
            fconfig: dict[str, Any],
    ) -> tuple[Optional[Any], bool]:
        if isinstance(
                field, (ForeignKeyFieldInstance, OneToOneFieldInstance, BackwardOneToOneRelation)
        ):
            return self._process_single_field_relation(field_name, field, json_schema_extra), True
        elif isinstance(field, (BackwardFKRelation, ManyToManyFieldInstance)):
            return self._process_many_field_relation(field_name, field), False
        elif field.field_type is JSONField:
            return Any, False
        return self._process_data_field(field_name, field, json_schema_extra, fconfig), False

    def _process_single_field_relation(
            self,
            field_name: str,
            field: Union[ForeignKeyFieldInstance, OneToOneFieldInstance, BackwardOneToOneRelation],
            json_schema_extra: dict[str, Any],
    ) -> Optional[Type[PydanticModel]]:
        python_type = getattr(field, "related_model", field.field_type)
        model: Optional[Type[PydanticModel]] = self._get_submodel(python_type, field_name)
        if model:
            self._relational_fields_index.append((field_name, model.__name__))
            if field.null:
                json_schema_extra["nullable"] = True
            if field.null or field.default is not None:
                model = Optional[model]  # type: ignore

            return model
        return None

    def _process_many_field_relation(
            self,
            field_name: str,
            field: Union[BackwardFKRelation, ManyToManyFieldInstance],
    ) -> Optional[Type[list[Type[PydanticModel]]]]:
        python_type = field.related_model
        model = self._get_submodel(python_type, field_name)
        if model:
            self._relational_fields_index.append((field_name, model.__name__))
            return list[model]  # type: ignore
        return None

    def _process_data_field(
            self,
            field_name: str,
            field: Field,
            json_schema_extra: dict[str, Any],
            fconfig: dict[str, Any],
    ) -> Optional[Any]:
        annotation = self._annotations.get(field_name, None)
        constraints = copy(field.constraints)
        if "readOnly" in constraints:
            json_schema_extra["readOnly"] = constraints["readOnly"]
            del constraints["readOnly"]
        fconfig.update(constraints)
        python_type: Union[Type[Enum], Type[IntEnum], Type]
        if isinstance(field, (IntEnumFieldInstance, CharEnumFieldInstance)):
            python_type = field.enum_type
        else:
            python_type = getattr(field, "related_model", field.field_type)
        ptype = python_type
        if field.null:
            json_schema_extra["nullable"] = True
        if not field.pk and (
                field_name in self._optional or field.default is not None or field.null
        ):
            ptype = Optional[ptype]
        if not (self._exclude_read_only and json_schema_extra.get("readOnly") is True):
            return annotation or ptype
        return None

    def _process_computed_field(
            self,
            field: ComputedFieldDescription,
    ) -> Optional[Any]:
        func = field.function
        annotation = get_annotations(self._cls, func).get("return", None)
        comment = _cleandoc(func)
        if annotation is not None:
            c_f = computed_field(return_type=annotation, description=comment)
            ret = c_f(func)
            return ret
        return None

    def _get_submodel(
            self, _model: Optional["Type[Model]"], field_name: str
    ) -> Optional[Type[PydanticModel]]:
        """Get Pydantic model for the submodel"""

        if _model:
            new_stack = self._stack + ((self._cls, field_name, self.meta.max_recursion),)

            # Get pydantic schema for the submodel
            prefix_len = len(field_name) + 1

            def get_fields_to_carry_on(field_tuple: tuple[str, ...]) -> tuple[str, ...]:
                return tuple(
                    str(v[prefix_len:]) for v in field_tuple if v.startswith(field_name + ".")
                )

            pmodel = _pydantic_recursion_protector(
                _model,
                exclude=get_fields_to_carry_on(self.meta.exclude),
                include=get_fields_to_carry_on(self.meta.include),
                computed=get_fields_to_carry_on(self.meta.computed),
                stack=new_stack,
                allow_cycles=self.meta.allow_cycles,
                sort_alphabetically=self.meta.sort_alphabetically,
            )
        else:
            pmodel = None

        # If the result is None it has been excluded and we need to exclude the field
        if pmodel is None:
            self.meta.exclude += (field_name,)

        return pmodel


def pydantic_model_creator(
        cls: "Type[Model]",
        *,
        name=None,
        exclude: Optional[tuple[str, ...]] = None,
        include: Optional[tuple[str, ...]] = None,
        computed: Optional[tuple[str, ...]] = None,
        optional: Optional[tuple[str, ...]] = None,
        allow_cycles: Optional[bool] = None,
        sort_alphabetically: Optional[bool] = None,
        exclude_readonly: bool = False,
        meta_override: Optional[Type] = None,
        model_config: Optional[ConfigDict] = None,
        validators: Optional[dict[str, Any]] = None,
        module: str = __name__,
) -> Type[PydanticModel]:
    """
    Function to build `Pydantic Model <https://docs.pydantic.dev/latest/concepts/models/>`__ off Tortoise Model.

    :param cls: The Tortoise Model
    :param name: Specify a custom name explicitly, instead of a generated name.
    :param exclude: Extra fields to exclude from the provided model.
    :param include: Extra fields to include from the provided model.
    :param computed: Extra computed fields to include from the provided model.
    :param optional: Extra optional fields for the provided model.
    :param allow_cycles: Do we allow any cycles in the generated model?
        This is only useful for recursive/self-referential models.

        A value of ``False`` (the default) will prevent any and all backtracking.
    :param sort_alphabetically: Sort the parameters alphabetically instead of Field-definition order.

        The default order would be:

            * Field definition order +
            * order of reverse relations (as discovered) +
            * order of computed functions (as provided).
    :param exclude_readonly: Build a subset model that excludes any readonly fields
    :param meta_override: A PydanticMeta class to override model's values.
    :param model_config: A custom config to use as pydantic config.
    :param validators: A dictionary of methods that validate fields.
    :param module: The name of the module that the model belongs to.

        Note: Created pydantic model uses config_class parameter and PydanticMeta's
            config_class as its Config class's bases(Only if provided!), but it
            ignores ``fields`` config. pydantic_model_creator will generate fields by
            include/exclude/computed parameters automatically.
    """
    pmc = PydanticModelCreator(
        cls=cls,
        name=name,
        exclude=exclude,
        include=include,
        computed=computed,
        optional=optional,
        allow_cycles=allow_cycles,
        sort_alphabetically=sort_alphabetically,
        exclude_readonly=exclude_readonly,
        meta_override=meta_override,
        model_config=model_config,
        validators=validators,
        module=module,
    )
    return pmc.create_pydantic_model()


__all__ = ['pydantic_model_creator']



"""
缓存测试
"""
import pytest
from redis import Redis
from datetime import datetime

from fastapi_tortoise_crud.cache import (
    CacheManager,
    cache,
    invalidate_cache,
    JsonEncoder
)


@pytest.fixture
def redis_client():
    """Redis客户端"""
    return Redis(host='localhost', port=6379, db=0)


@pytest.fixture
def cache_manager(redis_client):
    """缓存管理器"""
    return CacheManager(redis_client)


def test_json_encoder():
    """测试JSON编码器"""
    # 测试datetime编码
    now = datetime.now()
    encoder = JsonEncoder()
    encoded = encoder.default(now)
    assert isinstance(encoded, str)
    assert encoded == now.isoformat()


@pytest.mark.asyncio
async def test_cache_manager(cache_manager):
    """测试缓存管理器"""
    # 测试设置缓存
    key = "test_key"
    value = {"name": "test", "value": 123}
    await cache_manager.set(key, value, timeout=60)
    
    # 测试获取缓存
    cached = await cache_manager.get(key)
    assert cached == value
    
    # 测试删除缓存
    await cache_manager.delete(key)
    cached = await cache_manager.get(key)
    assert cached is None


@pytest.mark.asyncio
async def test_cache_decorator(cache_manager):
    """测试缓存装饰器"""
    call_count = 0
    
    @cache(timeout=60)
    async def test_func(param):
        nonlocal call_count
        call_count += 1
        return {"param": param, "count": call_count}
    
    # 第一次调用，应该执行函数
    result1 = await test_func("test")
    assert result1["count"] == 1
    
    # 第二次调用，应该从缓存获取
    result2 = await test_func("test")
    assert result2["count"] == 1  # count没有增加，说明使用了缓存
    assert result2 == result1


@pytest.mark.asyncio
async def test_invalidate_cache(cache_manager):
    """测试缓存失效装饰器"""
    # 设置一些测试缓存
    await cache_manager.set("test:1", "value1")
    await cache_manager.set("test:2", "value2")
    
    @invalidate_cache(key_pattern="test:*")
    async def clear_cache():
        return "cleared"
    
    # 执行函数，应该清除所有test:*的缓存
    await clear_cache()
    
    # 验证缓存已被清除
    assert await cache_manager.get("test:1") is None
    assert await cache_manager.get("test:2") is None


@pytest.mark.asyncio
async def test_cache_with_custom_key(cache_manager):
    """测试使用自定义键的缓存"""
    def key_builder(*args, **kwargs):
        return f"custom:{args[0]}:{kwargs.get('type', 'default')}"
    
    @cache(key_builder=key_builder)
    async def test_func(name, type="default"):
        return {"name": name, "type": type}
    
    # 使用不同参数调用
    result1 = await test_func("test", type="type1")
    result2 = await test_func("test", type="type2")
    
    # 验证缓存是否正确存储
    cached1 = await cache_manager.get("custom:test:type1")
    cached2 = await cache_manager.get("custom:test:type2")
    
    assert cached1 == result1
    assert cached2 == result2


@pytest.mark.asyncio
async def test_cache_clear_pattern(cache_manager):
    """测试按模式清除缓存"""
    # 设置不同模式的缓存
    await cache_manager.set("test:1", "value1")
    await cache_manager.set("test:2", "value2")
    await cache_manager.set("other:1", "value3")
    
    # 清除特定模式的缓存
    await cache_manager.clear("test:*")
    
    # 验证清除结果
    assert await cache_manager.get("test:1") is None
    assert await cache_manager.get("test:2") is None
    assert await cache_manager.get("other:1") is not None  # 不应被清除 


from typing import List, Callable
from fastapi import Depends, Query, HTTPException
from tortoise.contrib.pydantic import PydanticModel
from fastapi_pagination import Params
from fastapi_pagination.ext.tortoise import paginate

from ._base import CrudGenerator
from ._type import BaseApiOut


class ModelCrud(CrudGenerator):
    @classmethod
    async def pre_create(cls, item: PydanticModel) -> dict:
        return item.model_dump(exclude_unset=True)

    @classmethod
    async def pre_create_all(cls, items: List[PydanticModel]):
        for item in items:
            yield await cls.pre_create(item)

    @classmethod
    async def pre_update(cls, item: PydanticModel, item_id: str | int) -> dict:
        return item.model_dump(exclude_unset=True)

    @classmethod
    async def pre_list(cls, item: PydanticModel) -> dict:
        """
        数据预处理：搜索字段
        :param item:
        :return:
        """
        data = {}
        for k, v in item.model_dump(exclude_unset=True).items():
            # 如果v有值或者为bool类型
            if v or isinstance(v, bool):
                # 如果v为字符串并且有值，则使用模糊搜索
                if isinstance(v, str):
                    data[f'{k}__icontains'] = v
                # 单独处理时间字段
                elif k in ('create_time', 'update_time',):
                    data[f'{k}__range'] = v
                # 否则使用精确搜索
                else:
                    data[k] = v
        return data

    def route_list(self) -> Callable:
        schema_filters = self.schema_filters

        async def route(filters: schema_filters, params: Params = Depends(), order_by: str = '-create_time'):
            filter_item = await self.pre_list(filters)
            queryset = self.model.filter(**filter_item)
            if order_by:
                queryset = queryset.order_by(*order_by.split(','))
            data = await paginate(queryset, params, True)
            return BaseApiOut(data=data)

        return route

    def route_read(self) -> Callable:
        async def route(id: int = Query(..., description='id')):
            data = await self.model.find_one(id=id)
            if not data:
                raise HTTPException(status_code=404, detail=f'id：{id} 不存在')
            data = await self.schema_read.from_tortoise_orm(data)
            # data = await self.schema_read.from_orm(data)

            return BaseApiOut(data=data)

        return route

    def route_create(self) -> Callable:
        schema_create = self.schema_create

        async def route(item: schema_create):
            item = await self.pre_create(item)
            await self.model.create_one(item)
            return BaseApiOut()

        return route

    def route_create_all(self) -> Callable:
        schema_create = self.schema_create

        async def route(items: List[schema_create]):
            await self.model.bulk_create([self.model(**item) async for item in self.pre_create_all(items)],
                                         ignore_conflicts=False)
            return BaseApiOut(message='批量创建成功')

        return route

    def route_update(self) -> Callable:
        schema_update = self.schema_update

        async def route(id: int = Query(..., description='id'), item: schema_update = {}):
            item = await self.pre_update(item, item_id=id)
            data = await self.model.update_one(id, item)
            return BaseApiOut(data=item)

        return route

    def route_delete(self) -> Callable:
        async def route(ids: str = Query(..., description='item_ids')):
            data = await self.model.delete_many(ids.split(','))
            return BaseApiOut(data=data)

        return route

    def route_delete_all(self) -> Callable:
        async def route():
            await self.model.all().delete()
            return BaseApiOut(message='删除所有数据成功')

        return route


__all__ = [
    'ModelCrud'
]



from abc import ABC, abstractmethod
from typing import Union, Any, List, Callable, Type

from fastapi import APIRouter
from fastapi.types import DecoratedCallable
from fastapi_pagination import Page
# from pydantic import BaseModel as PydanticModel
from tortoise.contrib.pydantic.base import PydanticModel
from .model import BaseModel
from ._type import BaseApiOut, DEPENDENCIES


class CrudGenerator(APIRouter, ABC):
    def __init__(self, model: Union[BaseModel, Any],
                 schema_create: Union[bool, Type[PydanticModel]] = True,
                 schema_list: Union[bool, Type[PydanticModel]] = True,
                 schema_read: Union[bool, Type[PydanticModel]] = True,
                 schema_update: Union[bool, Type[PydanticModel]] = True,
                 schema_delete: Union[bool, Type[PydanticModel]] = True,
                 schema_filters: Union[bool, Type[PydanticModel]] = False,
                 dependencies: DEPENDENCIES = None,
                 override_dependencies: bool = True,
                 depends_read: Union[bool, DEPENDENCIES] = True,
                 depends_create: Union[bool, DEPENDENCIES] = True,
                 depends_update: Union[bool, DEPENDENCIES] = True,
                 depends_delete: Union[bool, DEPENDENCIES] = True,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.model = model

        self.dependencies = dependencies or []
        self.override_dependencies = override_dependencies
        
        def validate_schema(schema, model_method, fallback_schema=None):
            if isinstance(schema, type):
                if not issubclass(schema, PydanticModel):
                    raise TypeError(f"Schema must be a subclass of PydanticModel, got {schema}")
                return schema
            if fallback_schema:
                return fallback_schema
            return getattr(model, model_method)()

        self.schema_read = validate_schema(schema_read, 'schema_read')
        self.schema_list = validate_schema(schema_list, 'schema_list', self.schema_read)
        self.schema_update = validate_schema(schema_update, 'schema_update')
        self.schema_create = validate_schema(schema_create, 'schema_create')
        self.schema_delete = validate_schema(schema_delete, 'schema_delete')
        
        if schema_filters:
            self.schema_filters = validate_schema(schema_filters, 'schema_filters')
        else:
            self.schema_filters = model.schema_filters()

        model_name = model.__name__.lower()
        model_title = getattr(model.Meta, 'table_title', model_name)
        if schema_list:
            self.add_api_route(
                '/list',
                self.route_list(),
                methods=['POST'],
                response_model=BaseApiOut[Page[self.schema_list]],
                name=f'{model_name}Read',
                summary=f'{model_title}列表',
                dependencies=depends_read
            )
        if schema_read:
            self.add_api_route(
                '/read',
                self.route_read(),
                methods=['GET'],
                response_model=BaseApiOut[self.schema_read],
                name=f'{model_name}Read',
                summary=f'{model_title}查看',
                dependencies=depends_read
            )

        if self.schema_create:
            self.add_api_route(
                '/create',
                self.route_create(),
                methods=['POST'],
                response_model=BaseApiOut,
                name=f'{model_name}Create',
                summary=f'{model_title}创建',
                dependencies=depends_create
            )

            self.add_api_route(
                '/createall',
                self.route_create_all(),
                methods=['POST'],
                response_model=BaseApiOut,
                name=f'{model_name}Create',
                summary=f'{model_title}创建所有',
                dependencies=depends_create
            )
        if self.schema_update:
            self.add_api_route(
                '/update',
                self.route_update(),
                methods=['PUT'],
                response_model=BaseApiOut,
                name=f'{model_name}Update',
                summary=f'{model_title}更新',
                dependencies=depends_update
            )
        if self.schema_delete:
            self.add_api_route(
                '/delete',
                self.route_delete(),
                methods=['DELETE'],
                response_model=BaseApiOut,
                description='删除1条或多条数据example：1,2',
                name=f'{model_name}Delete',
                summary=f'{model_title}删除',
                dependencies=depends_delete
            )
            self.add_api_route(
                '/deleteall',
                self.route_delete_all(),
                methods=['DELETE'],
                response_model=BaseApiOut,
                description='删除所有数据',
                name=f'{model_name}Delete',
                summary=f'{model_title}删除所有',
                dependencies=depends_delete
            )

    def add_api_route(
            self,
            path: str,
            endpoint: Callable[..., Any],
            dependencies: Union[bool, DEPENDENCIES],
            *args,
            **kwargs: Any,
    ) -> None:
        # bool类型获取None都设置为空列表
        new_dependencies = [] if isinstance(dependencies, bool) or dependencies is None else dependencies
        if self.override_dependencies:
            original_dependencies = self.dependencies.copy()

            if dependencies is False or (isinstance(dependencies, list) and len(dependencies) == 0):
                self.dependencies = []
                try:
                    super().add_api_route(path, endpoint, dependencies=new_dependencies, **kwargs)
                finally:
                    self.dependencies = original_dependencies
                    return
        super().add_api_route(path, endpoint, dependencies=new_dependencies, **kwargs)

    def api_route(
            self, path: str, *args: Any, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """Overrides and exiting route if it exists"""
        methods = kwargs["methods"] if "methods" in kwargs else ["GET"]
        self.remove_api_route(path, methods)
        return super().api_route(path, *args, **kwargs)

    def get(
            self, path: str, *args: Any, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        self.remove_api_route(path, ["Get"])
        return super().get(path, *args, **kwargs)

    def post(
            self, path: str, *args: Any, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        self.remove_api_route(path, ["POST"])
        return super().post(path, *args, **kwargs)

    def put(
            self, path: str, *args: Any, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        self.remove_api_route(path, ["PUT"])
        return super().put(path, *args, **kwargs)

    def delete(
            self, path: str, *args: Any, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        self.remove_api_route(path, ["DELETE"])
        return super().delete(path, *args, **kwargs)

    def remove_api_route(self, path: str, methods: List[str]) -> None:
        methods_ = set(methods)

        for route in self.routes:
            if (
                    route.path == f"{self.prefix}{path}"  # type: ignore
                    and route.methods == methods_  # type: ignore
            ):
                self.routes.remove(route)

    @abstractmethod
    def route_list(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_read(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_update(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_create(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_create_all(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_delete(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError

    @abstractmethod
    def route_delete_all(self, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        raise NotImplementedError


__all__ = [
    'CrudGenerator'
]



"""
模型测试
"""
import pytest
from tortoise import Tortoise

from fastapi_tortoise_crud import BaseModel
from fastapi_tortoise_crud.model import fields


# 测试模型
class TestModel(BaseModel):
    name = fields.CharField(max_length=50, description="名称")
    description = fields.TextField(null=True, description="描述")
    
    class Meta:
        table = "test_models"


@pytest.fixture
async def init_db():
    """初始化数据库"""
    await Tortoise.init(
        db_url='sqlite://:memory:',
        modules={'models': ['tests.test_model']}
    )
    await Tortoise.generate_schemas()
    yield
    await Tortoise.close_connections()


@pytest.mark.asyncio
async def test_create_model(init_db):
    """测试创建模型"""
    # 创建记录
    model = await TestModel.create(name="test1", description="desc1")
    assert model.id is not None
    assert model.name == "test1"
    assert model.description == "desc1"
    assert model.status is True
    assert model.version == 1


@pytest.mark.asyncio
async def test_update_model(init_db):
    """测试更新模型"""
    # 创建记录
    model = await TestModel.create(name="test1")
    
    # 更新记录
    await TestModel.update_one(str(model.id), {"name": "test2"})
    
    # 验证更新
    updated = await TestModel.find_one(id=model.id)
    assert updated.name == "test2"
    assert updated.version == 2  # 版本号增加


@pytest.mark.asyncio
async def test_delete_model(init_db):
    """测试删除（禁用）"""
    # 创建记录
    model = await TestModel.create(name="test1")
    
    # 删除（禁用）记录
    await TestModel.delete_one(str(model.id))
    
    # 验证记录已被禁用
    deleted = await TestModel.get_or_none(id=model.id)
    assert deleted.status is False
    
    # 验证查询时会过滤掉已禁用的记录
    result = await TestModel.find_one(id=model.id)
    assert result is None


@pytest.mark.asyncio
async def test_find_with_status(init_db):
    """测试状态过滤"""
    # 创建记录
    model1 = await TestModel.create(name="test1")
    model2 = await TestModel.create(name="test2", status=False)
    
    # 测试默认只查询启用的记录
    results = await TestModel.find_by()
    assert len(results) == 1
    assert results[0].name == "test1"
    
    # 测试查询所有记录
    results = await TestModel.find_by(status=None)
    assert len(results) == 2
    
    # 测试只查询禁用的记录
    results = await TestModel.find_by(status=False)
    assert len(results) == 1
    assert results[0].name == "test2"


@pytest.mark.asyncio
async def test_metadata(init_db):
    """测试元数据"""
    # 创建带元数据的记录
    metadata = {"key1": "value1", "key2": "value2"}
    model = await TestModel.create(
        name="test1",
        metadata=metadata
    )
    
    # 验证元数据
    assert model.metadata == metadata
    
    # 更新元数据
    new_metadata = {"key3": "value3"}
    await TestModel.update_one(str(model.id), {"metadata": new_metadata})
    
    # 验证更新后的元数据
    updated = await TestModel.find_one(id=model.id)
    assert updated.metadata == new_metadata 


# -*- coding: utf-8 -*-
"""
@author: moxiaoying
@create: 2022/10/16
@description: 基础模型
"""
import asyncio
from datetime import datetime
from typing import Type, Optional, List

from tortoise import fields, Model
from pydantic import ConfigDict, Field
from .utils import pydantic_model_creator


class BaseCrudMixin(Model):
    @classmethod
    async def create_one(cls, item: dict):
        return await cls.create(**item)

    @classmethod
    async def find_by(cls, **kwargs):
        return await cls.filter(**kwargs).all()

    @classmethod
    async def find_one(cls, **kwargs):
        return await cls.filter(**kwargs).first()

    @classmethod
    async def update_one(cls, _id: str | int, item: dict):
        update_obj = await cls.get_or_none(id=_id)
        if not update_obj:
            return
        await update_obj.update_from_dict(item).save()
        return update_obj

    @classmethod
    async def delete_one(cls, _id: str) -> int:
        deleted_count = await cls.filter(id=_id).delete()
        return deleted_count

    @classmethod
    async def delete_many(cls, ids: list) -> int:
        deleted_count = await cls.filter(id__in=ids).delete()
        return deleted_count


class BaseSchemaMixin:
    @classmethod
    def base_schema(cls: Type[Model], name, include=(), exclude=(), **kwargs):
        name = f'{cls.__name__}Schema{name}'
        optional = kwargs.pop('optional', ())
        # optional = kwargs.pop('optional', ()) or cls._meta.fields
        # print(f'{name}\t{optional}')
        # include = kwargs.get('include', ())
        if include:
            a = pydantic_model_creator(cls, name=name, include=include, optional=optional,
                                       **kwargs)
            return a
        return pydantic_model_creator(cls, name=name, optional=optional, exclude=exclude,
                                      **kwargs)

    @classmethod
    def schema_list(cls, name='List', include=(), exclude=(), **kwargs):
        return cls.base_schema(name=name, include=include, exclude=exclude, **kwargs)

    @classmethod
    def schema_create(cls, name='Create', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True, **kwargs)

    @classmethod
    def schema_read(cls, name='Read', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, **kwargs)

    @classmethod
    def schema_update(cls, name='Update', include=(), exclude=(), **kwargs):
        return cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True,
                               **kwargs)

    @classmethod
    def schema_filters(cls, name='Filters', include=(), exclude=(), add_default=True, **kwargs):
        """
        :param name:
        :param include:
        :param exclude:
        :param add_default: 为True添加默认字段用于前端搜索（status、create_time、update_time） False走默认
        :param kwargs:
        :return:
        """
        # exclude如果有status、create_time、update_time就删除掉，include如果没有就添加
        if not add_default:
            return cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True,
                                   **kwargs)
        include = tuple(set(include) | {'status'})
        exclude = tuple(set(exclude) - {'status'})
        filter_schema = cls.base_schema(name, include=include, exclude=exclude, exclude_readonly=True,
                                        **kwargs)

        class FilterSchema(filter_schema):
            create_time: Optional[List[datetime]] = Field(None, description="创建时间")
            update_time: Optional[List[datetime]] = Field(None, description="更新时间")

        return FilterSchema

    @classmethod
    def schema_delete(cls):
        return int


class TimestampMixin:
    create_time = fields.DatetimeField(
        null=True, auto_now_add=True, description="创建时间")
    update_time = fields.DatetimeField(
        null=True, auto_now=True, description="更新时间")


class BaseModel(BaseCrudMixin, BaseSchemaMixin, TimestampMixin):
    id = fields.IntField(pk=True, index=True, description="主键")
    status = fields.BooleanField(
        null=False, default=True, index=True, description="状态:True=启用,False=禁用")

    async def to_dict(self, m2m: bool = False, exclude_fields: list[str] | None = None):
        if exclude_fields is None:
            exclude_fields = []

        d = {}
        for field in self._meta.db_fields:
            if field not in exclude_fields:
                value = getattr(self, field)
                if isinstance(value, datetime):
                    value = value.strftime('%Y-%m-%d %H:%M:%S')
                d[field] = value
        if m2m:
            tasks = [self.__fetch_m2m_field(field) for field in self._meta.m2m_fields if field not in exclude_fields]
            results = await asyncio.gather(*tasks)
            for field, values in results:
                d[field] = values
        return d

    async def __fetch_m2m_field(self, field):
        values = [value for value in await getattr(self, field).all().values()]
        for value in values:
            value.update((k, v.strftime('%Y-%m-%d %H:%M:%S')) for k, v in value.items() if isinstance(v, datetime))
        return field, values

    class PydanticMeta:
        backward_relations = False
        model_config = ConfigDict(extra='ignore', strict=False)

    class Meta:
        abstract = True
        ordering = ['-update_time', '-create_time']


__all__ = [
    'BaseModel'
]



"""
批量操作测试
"""
import pytest
from tortoise import Tortoise

from fastapi_tortoise_crud import BaseModel
from fastapi_tortoise_crud.bulk import BulkOperations
from fastapi_tortoise_crud.model import fields


# 测试模型
class BulkTestModel(BaseModel):
    name = fields.CharField(max_length=50, description="名称")
    value = fields.IntField(description="值")
    
    class Meta:
        table = "bulk_test_models"


@pytest.fixture
async def init_db():
    """初始化数据库"""
    await Tortoise.init(
        db_url='sqlite://:memory:',
        modules={'models': ['tests.test_bulk']}
    )
    await Tortoise.generate_schemas()
    yield
    await Tortoise.close_connections()


@pytest.mark.asyncio
async def test_bulk_create(init_db):
    """测试批量创建"""
    # 准备测试数据
    items = [
        {"name": f"test{i}", "value": i}
        for i in range(5)
    ]
    
    # 批量创建
    results = await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 验证结果
    assert len(results) == 5
    for i, result in enumerate(results):
        assert result.name == f"test{i}"
        assert result.value == i
        assert result.status is True
        assert result.version == 1


@pytest.mark.asyncio
async def test_bulk_update(init_db):
    """测试批量更新"""
    # 创建测试数据
    items = [
        {"name": f"test{i}", "value": i}
        for i in range(5)
    ]
    original_models = await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 准备更新数据
    updates = [
        {
            "id": model.id,
            "value": model.value + 10
        }
        for model in original_models
    ]
    
    # 批量更新
    updated_count = await BulkOperations.bulk_update(
        BulkTestModel,
        updates,
        fields=['value']
    )
    
    # 验证结果
    assert updated_count == 5
    updated_models = await BulkTestModel.all()
    for i, model in enumerate(updated_models):
        assert model.value == i + 10
        assert model.version == 2  # 版本号应该增加


@pytest.mark.asyncio
async def test_bulk_delete(init_db):
    """测试批量删除（禁用）"""
    # 创建测试数据
    items = [
        {"name": f"test{i}", "value": i}
        for i in range(5)
    ]
    models = await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 获取ID列表
    ids = [model.id for model in models]
    
    # 批量禁用
    deleted_count = await BulkOperations.bulk_delete(
        BulkTestModel,
        {'id__in': ids}
    )
    
    # 验证结果
    assert deleted_count == 5
    
    # 验证记录已被禁用
    for model_id in ids:
        model = await BulkTestModel.get_or_none(id=model_id)
        assert model.status is False
    
    # 验证查询时会过滤掉已禁用的记录
    remaining = await BulkTestModel.find_by()
    assert len(remaining) == 0


@pytest.mark.asyncio
async def test_bulk_physical_delete(init_db):
    """测试批量物理删除"""
    # 创建测试数据
    items = [
        {"name": f"test{i}", "value": i}
        for i in range(5)
    ]
    models = await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 获取ID列表
    ids = [model.id for model in models]
    
    # 批量物理删除
    deleted_count = await BulkOperations.bulk_delete(
        BulkTestModel,
        {'id__in': ids},
        physical_delete=True
    )
    
    # 验证结果
    assert deleted_count == 5
    
    # 验证记录已被物理删除
    for model_id in ids:
        model = await BulkTestModel.get_or_none(id=model_id)
        assert model is None


@pytest.mark.asyncio
async def test_bulk_get(init_db):
    """测试批量获取"""
    # 创建测试数据
    items = [
        {"name": f"test{i}", "value": i, "status": i % 2 == 0}
        for i in range(5)
    ]
    original_models = await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 获取ID列表
    ids = [model.id for model in original_models]
    
    # 测试只获取启用的记录
    results = await BulkOperations.bulk_get(BulkTestModel, ids)
    assert len(results) == 3  # 只有status为True的记录
    
    # 测试获取所有记录
    results = await BulkOperations.bulk_get(BulkTestModel, ids, include_disabled=True)
    assert len(results) == 5  # 所有记录


@pytest.mark.asyncio
async def test_bulk_create_with_conflicts(init_db):
    """测试带冲突处理的批量创建"""
    # 创建初始数据
    items = [
        {"name": "test1", "value": 1},
        {"name": "test2", "value": 2}
    ]
    await BulkOperations.bulk_create(BulkTestModel, items)
    
    # 尝试创建包含冲突的数据
    conflict_items = [
        {"name": "test1", "value": 3},  # 冲突
        {"name": "test3", "value": 3}   # 新数据
    ]
    
    # 使用ignore_conflicts=True
    results = await BulkOperations.bulk_create(
        BulkTestModel,
        conflict_items,
        ignore_conflicts=True
    )
    
    # 验证结果
    all_models = await BulkTestModel.all()
    assert len(all_models) == 3  # 应该有3条记录
    
    # 验证值没有被覆盖
    model = await BulkTestModel.get(name="test1")
    assert model.value == 1  # 原始值应该保持不变 